pkt_header = ('>II', ['type', 'length'])
pkt_types = {2: [('>IIIIIII',
      ['sensor_id',
       'event_id',
       'event_second',
       'packet_second',
       'packet_microsecond',
       'linktype',
       'packet_length'])],
 7: [('>IIIIIIIIIIIHHBBBBIHH',
      ['sensor_id',
       'event_id',
       'event_second',
       'event_microsecond',
       'signature_id',
       'generator_id',
       'signature_revision',
       'classification_id',
       'priority_id',
       'ip_source',
       'ip_destination',
       'sport_itype',
       'dport_icode',
       'protocol',
       'impact_flag',
       'impact',
       'blocked',
       'mpls_label',
       'vlanId',
       'pad2']),
     ('>IIIIIIIIIIIHHBBBB',
      ['sensor_id',
       'event_id',
       'event_second',
       'event_microsecond',
       'signature_id',
       'generator_id',
       'signature_revision',
       'classification_id',
       'priority_id',
       'ip_source',
       'ip_destination',
       'sport_itype',
       'dport_icode',
       'protocol',
       'impact_flag',
       'impact',
       'blocked'])],
 72: [('>IIIIIIIIIIIIIIIIIHHBBBBIHH',
       ['sensor_id',
        'event_id',
        'event_second',
        'event_microsecond',
        'signature_id',
        'generator_id',
        'signature_revision',
        'classification_id',
        'priority_id',
        'ip_source',
        'ip_source',
        'ip_source',
        'ip_source',
        'ip_destination',
        'ip_destination',
        'ip_destination',
        'ip_destination',
        'sport_itype',
        'dport_icode',
        'protocol',
        'impact_flag',
        'impact',
        'blocked',
        'mpls_label',
        'vlanId',
        'pad2']),
      ('>IIIIIIIIIIIIIIIIIHHBBBB',
       ['sensor_id',
        'event_id',
        'event_second',
        'event_microsecond',
        'signature_id',
        'generator_id',
        'signature_revision',
        'classification_id',
        'priority_id',
        'ip_source',
        'ip_source',
        'ip_source',
        'ip_source',
        'ip_destination',
        'ip_destination',
        'ip_destination',
        'ip_destination',
        'sport_itype',
        'dport_icode',
        'protocol',
        'impact_flag',
        'impact',
        'blocked'])],
 104: [('>IIIIIIIIIIIHHBBBBIHH',
        ['sensor_id',
         'event_id',
         'event_second',
         'event_microsecond',
         'signature_id',
         'generator_id',
         'signature_revision',
         'classification_id',
         'priority_id',
         'ip_source',
         'ip_destination',
         'sport_itype',
         'dport_icode',
         'protocol',
         'impact_flag',
         'impact',
         'blocked',
         'mpls_label',
         'vlanId',
         'pad2'])],
 105: [('>IIIIIIIIIIIIIIIIIHHBBBBIHH',
        ['sensor_id',
         'event_id',
         'event_second',
         'event_microsecond',
         'signature_id',
         'generator_id',
         'signature_revision',
         'classification_id',
         'priority_id',
         'ip_source',
         'ip_source',
         'ip_source',
         'ip_source',
         'ip_destination',
         'ip_destination',
         'ip_destination',
         'ip_destination',
         'sport_itype',
         'dport_icode',
         'protocol',
         'impact_flag',
         'impact',
         'blocked',
         'mpls_label',
         'vlanId',
         'pad2'])],
 110: [('>II', ['event_type', 'event_length'])],
 207: [('>IIIIIIIIIIIHHBBBBIHH16sIIIII16s16s16s16s16s',
        ['sensor_id',
         'event_id',
         'event_second',
         'event_microsecond',
         'signature_id',
         'generator_id',
         'signature_revision',
         'classification_id',
         'priority_id',
         'ip_source',
         'ip_destination',
         'sport_itype',
         'dport_icode',
         'protocol',
         'impact_flag',
         'impact',
         'blocked',
         'mpls_label',
         'vlanId',
         'pad',
         'policy_uuid',
         'user_id',
         'web_application_id',
         'client_application_id',
         'application_protocol_id',
         'policyengine_rule_id',
         'policyengine_policy_uuid',
         'interface_ingress_uuid',
         'interface_egress_uuid',
         'security_zone_ingress_uuid',
         'security_zone_egress_uuid'])],
 208: [('>IIIIIIIIIIIIIIIIIHHBBBBIHH16sIIIII16s16s16s16s16s',
        ['sensor_id',
         'event_id',
         'event_second',
         'event_microsecond',
         'signature_id',
         'generator_id',
         'signature_revision',
         'classification_id',
         'priority_id',
         'ip_source',
         'ip_source',
         'ip_source',
         'ip_source',
         'ip_destination',
         'ip_destination',
         'ip_destination',
         'ip_destination',
         'sport_itype',
         'dport_icode',
         'protocol',
         'impact_flag',
         'impact',
         'blocked',
         'mpls_label',
         'vlanId',
         'pad',
         'policy_uuid',
         'user_id',
         'web_application_id',
         'client_application_id',
         'application_protocol_id',
         'policyengine_rule_id',
         'policyengine_policy_uuid',
         'interface_ingress_uuid',
         'interface_egress_uuid',
         'security_zone_ingress_uuid',
         'security_zone_egress_uuid'])]}
pkt_tails = {110: [('>IIIIII',
        ['sensor_id',
         'event_id',
         'event_second',
         'type',
         'data_type',
         'blob_length'])]}


if __name__ == '__main__':
	# Usage:
	#  bzr branch lp:pyclibrary
	#  cd pyclibrary
	#  python .../_snort_u2_format.py .../snort-2.9.2.1/src/sfutil/Unified2_common.h

	import sys, struct, CParser

	parser = CParser.CParser(sys.argv[1:])
	parser.processAll()

	structs = dict(
		(name, parser.defs['structs'][k]) for name, (ctype, k) in (
			(name, ( typedef if isinstance(typedef, tuple)
				else tuple(typedef[0].split(None, 1)) ))
			for name, typedef in parser.defs['types'].viewitems()
			if 'Unified2' in name )
		if ctype == 'struct' and ' ' not in name )

	def struct_spec( name,
			_ctypes={ 'uint32_t': 'I', 'uint16_t': 'H',
				'uint8_t': 'B', 'struct in6_addr': 'IIII' } ):
		typedef, names = '', list()
		for name, ctype, unused in structs[name]['members']:
			if len(ctype) == 2: ctype, cnt = ctype[0], ctype[1][0]
			elif len(ctype) == 1: ctype, cnt = ctype[0], 1
			else: raise ValueError(ctype)
			try:
				spec_atom = _ctypes[ctype]
				spec, cnt = spec_atom * cnt, len(spec_atom) * cnt
			except KeyError: print 'K', ctype
			if cnt > 1 and spec_atom == 'B':
				spec, cnt = '{}s'.format(cnt), 1
			for i in xrange(cnt):
				typedef += spec
				names.append(name)
		return '>' + typedef, names

	types = dict( (k,v)
		for k,v in parser.defs['values'].viewitems()
		if k.startswith('UNIFIED2_') )
	pkt_types = dict(
		( types[k], [struct_spec(v)]
			if isinstance(v, bytes) else map(struct_spec, v) )
		for k,v in dict(
			UNIFIED2_IDS_EVENT=[
				'Unified2IDSEvent', 'Serial_Unified2IDSEvent_legacy' ],
			UNIFIED2_IDS_EVENT_IPV6=[
				'Unified2IDSEventIPv6', 'Serial_Unified2IDSEventIPv6_legacy' ],
			UNIFIED2_IDS_EVENT_NG='Unified2IDSEventNG',
			UNIFIED2_IDS_EVENT_IPV6_NG='Unified2IDSEventIPv6_NG',
			UNIFIED2_PACKET='Serial_Unified2Packet',
			UNIFIED2_EXTRA_DATA='Unified2ExtraDataHdr',
			UNIFIED2_IDS_EVENT_VLAN='Unified2IDSEvent',
			UNIFIED2_IDS_EVENT_IPV6_VLAN='Unified2IDSEventIPv6' ).viewitems() )
	pkt_tails = dict(
		( types[k], [struct_spec(v)]
			if isinstance(v, bytes) else map(struct_spec, v) )
		for k,v in dict(
			UNIFIED2_EXTRA_DATA='SerialUnified2ExtraData' ).viewitems() )
	pkt_header = struct_spec('Serial_Unified2_Header')

	from pprint import pformat
	print 'pkt_header = {}'.format(pformat(pkt_header))
	print 'pkt_types = {}'.format(pformat(pkt_types))
	print 'pkt_tails = {}'.format(pformat(pkt_tails))



# -*- coding: utf-8 -*-
from __future__ import print_function

import itertools as it, operator as op, functools as ft
import struct

from unified2._format import pkt_header, pkt_types, pkt_tails


class Parser(object):

	def __init__(self):
		self.data_fmts = dict(header=self.struct(pkt_header))
		self.data_fmts.update((k, map(self.struct, v)) for k,v in pkt_types.viewitems())
		self.data_tails = dict((k, map(self.struct, v)) for k,v in pkt_tails.viewitems())
		self.read_fmt = 'header'
		self.read_pos, self.read_len = 0, self.data_fmts[self.read_fmt][0]

	def struct(self, spec):
		fmt, keys = spec
		return struct.calcsize(fmt), fmt, keys

	def unpack(self, type_id, buff, type_tail=False):
		# Snort type_id=7,72 can match different structs of different length,
		#  hence the check by-length - longest fitting struct is chosen.
		# How anyone could ever think that it's a good idea? *sigh*
		pkt_fmt, pkt_len = ( self.data_fmts
			if not type_tail else self.data_tails )[type_id], len(buff)
		if isinstance(pkt_fmt, list):
			for fmt_len, fmt, keys in pkt_fmt:
				if fmt_len <= pkt_len: break
			else:
				raise ValueError('Packet length in header is'
					' lesser than size of all known structs for this type')
		else: fmt_len, fmt, keys = pkt_fmt

		pkt, tail = dict(), buff[fmt_len:]
		for k,v in it.izip(keys, struct.unpack_from(fmt, buff)):
			if k not in pkt: pkt[k] = v
			elif isinstance(pkt[k], list): pkt[k].append(v)
			else: pkt[k] = [pkt[k], v]

		if not type_tail and type_id in self.data_tails:
			tail = self.unpack(type_id, tail, type_tail=True)
		return pkt, tail

	def read(self, src):
		buff = src.read(self.read_len)
		self.read_pos += len(buff)
		return buff

	def process(self, buff):
		if self.read_pos < self.read_len: return buff, None
		buff, buff_tail = buff[:self.read_len], buff[self.read_len:]
		pkt, pkt_tail = self.unpack(self.read_fmt, buff)
		if self.read_fmt == 'header':
			self.read_len, self.read_fmt = pkt['length'], pkt['type']
			pkt = None
		else:
			self.read_fmt = 'header'
			self.read_len = self.data_fmts[self.read_fmt][0]
			pkt = pkt, pkt_tail
		self.read_pos = len(buff_tail)
		return buff_tail, pkt


def read(src):
	'Event generator from u2 stream.'
	parser, buff_agg = Parser(), ''
	while True:
		buff = parser.read(src)
		if not buff: break # EOF
		buff_agg += buff
		while True:
			buff_agg, ev = parser.process(buff_agg)
			if ev is None: break
			yield ev

def parse(path):
	'Event generator from path to u2 file.'
	return read(open(path, 'rb'))


if __name__ == '__main__':
	import sys
	for ev, ev_tail in parse(sys.argv[1]):
		print('Event:', ev)
		if ev_tail: print('Event tail:', ev_tail)



#!/usr/bin/env python

import os, sys

from setuptools import setup, find_packages

# Error-handling here is to allow package to be built w/o README included
try:
	readme = open(os.path.join(
		os.path.dirname(__file__), 'README.txt' )).read()
except IOError: readme = ''

setup(

	name = 'unified2',
	version = '12.07.0',
	author = 'Mike Kazantsev',
	author_email = 'mk.fraggod@gmail.com',
	license = 'WTFPL',
	keywords = 'unified2 u2 ids snort suricata parser',
	url = 'http://github.com/mk-fg/unified2',

	description = 'unified2 IDS binary log format parser',
	long_description = readme,

	classifiers = [
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Intended Audience :: System Administrators',
		'Intended Audience :: Telecommunications Industry',
		'License :: OSI Approved',
		'Operating System :: POSIX',
		'Operating System :: Unix',
		'Programming Language :: Python',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 2 :: Only',
		'Topic :: Security',
		'Topic :: System :: Networking :: Monitoring' ],

	packages = find_packages(),
	package_data = {'': ['README.txt']},
	exclude_package_data = {'': ['README.*']} )
